#include "plugins.h"
#include <QDebug>

Plugin::Plugin() :
		m_stop(false)
{
}

Plugin::~Plugin(){
}

//------------------------------------------------------------------------------
//begin of info part

QString VERSION_TO_STRING(quint32 version){
	quint8 build = (version & 0xFFFFFF);
	quint8 revision = (version & 0xFFFF) >> 8;
	quint8 minor = (version & 0xFFFFFF) >> 16;
	quint8 major = version >> 24;

	QString versionString = QString::number(major) + "." +
							QString::number(minor) + "." +
							QString::number(revision) + "." +
							QString::number(build);


							;
	return versionString;
}

void Plugin::setShortNameInfo(const QString &shortName){
	m_pluginInfo.shortName = shortName;
}

QString Plugin::getShortNameInfo() const{
	return m_pluginInfo.shortName;
}

void Plugin::setVersionInfo(const quint32 &version){
	m_pluginInfo.version = version;
	m_pluginInfo.versionString = VERSION_TO_STRING( version );
}

quint32 Plugin::getVersionInfo() const{
	return m_pluginInfo.version;
}

QString Plugin::getVersionStringInfo() const{
	return m_pluginInfo.versionString;
}

void Plugin::setDescriptionInfo(const QString &description){
	m_pluginInfo.description = description;
}

QString Plugin::getDescriptionInfo() const{
	return m_pluginInfo.description;
}

void Plugin::setAuthorInfo(const QString &author){
	m_pluginInfo.author = author;
}

QString Plugin::getAuthorInfo() const{
	return m_pluginInfo.author;
}

void Plugin::setAuthorEmailInfo(const QString &authorEmail){
	m_pluginInfo.authorEmail = authorEmail;
}

QString Plugin::getAuthorEmailInfo() const{
	return m_pluginInfo.authorEmail;
}

void Plugin::setCopyrightInfo(const QString &copyright){
	m_pluginInfo.copyright = copyright;
}

QString Plugin::getCopyrightInfo() const{
	return m_pluginInfo.copyright;
}

void Plugin::setHomepageInfo(const QString &homepage){
	m_pluginInfo.homepage = homepage;
}

QString Plugin::getHomepageInfo() const{
	return m_pluginInfo.homepage;
}

PLUGININFO Plugin::getPluginInfo() const {
	return m_pluginInfo;
}

void Plugin::setPluginInfo(const PLUGININFO &pluginInfo){
	m_pluginInfo = pluginInfo;
}

void Plugin::setPluginInfo(const QString &shortName,
						   const quint32 &version,
						   const QString &description,
						   const QString &author,
						   const QString &authorEmail,
						   const QString &copyright,
						   const QString &homepage)
{
	m_pluginInfo.shortName = shortName;
	m_pluginInfo.version = version;

	m_pluginInfo.versionString = VERSION_TO_STRING( version );

	m_pluginInfo.description = description;
	m_pluginInfo.author = author;
	m_pluginInfo.authorEmail = authorEmail;
	m_pluginInfo.copyright = copyright;
	m_pluginInfo.homepage = homepage;
}

//end of info part
//------------------------------------------------------------------------------


bool Plugin::awaitData(){
	// defines whether the encapsulating loop should issue a "continue"
	// statement, thus skipping the cycle
	bool shouldSkip = false;
#ifdef THREAD_DEBUG_OUTPUT
	qDebug() << "plugin start";
#endif
	m_pPluginMutex->lock();
	if ((*m_pNumUsedBytes) == 0) {
		m_pBufferNotEmpty->wait(m_pPluginMutex);
	}
	m_pPluginMutex->unlock();
	/* data is available, thread goes green */

#ifdef THREAD_DEBUG_OUTPUT
	qDebug() << "plugin start work";
#endif
	/* when thread wakes up, there not necessarily is new data available.
	   instead, some other thread might request us to stop. */
	if (m_stop) {
#ifdef THREAD_DEBUG_OUTPUT
		qDebug() << "plugin: woken up, exiting";
#endif
		shouldSkip = true;
	}
	return shouldSkip;
}

EventContainer *Plugin::getData() const
{
#ifdef THREAD_DEBUG_OUTPUT
	qDebug() << "plugin:" << *m_indexThread;
#endif
	EventContainer *event = m_pBuffer->at(*m_indexThread);
	//qDebug() << "plugin events:" << event->getEvents();
	*m_indexThread = (*m_indexThread + 1) % INPUT_BUFFER_SIZE;
	return event;
}

void Plugin::announceProcessingEnd(){
#ifdef THREAD_DEBUG_OUTPUT
	qDebug() << "plugin end work";
#endif

	/* announce finished work */
	m_pPluginMutex->lock();
	--(*m_pNumUsedBytes);
	m_pBufferNotFull->wakeAll();
	m_pPluginMutex->unlock();
	/* announcment done */
#ifdef THREAD_DEBUG_OUTPUT
	qDebug() << "plugin end";
#endif
}

void Plugin::waitForUndo()
{
	m_pUndoMutex->lock();
}

void Plugin::releaseUndo()
{
	m_pUndoMutex->unlock();
}

void Plugin::run(){
#ifdef THREAD_DEBUG_OUTPUT
	qDebug() << "plugin: start!";
#endif
	while (!m_stop) {
		//this function is divided up into sections with comments guarding start
		//and end of each.

		/* wait for data to become available */
		if (awaitData() == true) {
			// continue rather than return, so that cleanup can be done after
			// upper-level while() finishes.
			continue;
		}
		/* data is available, thread goes green */

		/* perform some threaded work on data */
		EventContainer *event = getData();
		onDataReceived(event);
		/* my work is done here */

		announceProcessingEnd();
	}
#ifdef THREAD_DEBUG_OUTPUT
	qDebug() << "plugin: run() has finished";
#endif
}

void Plugin::stop(){
#ifdef THREAD_DEBUG_OUTPUT
	qDebug() << "plugin: stopping thread";
#endif
	m_stop = true;
	m_pBufferNotEmpty->wakeAll();
#ifdef THREAD_DEBUG_OUTPUT
	qDebug() << "plugin: now run() has to exit";
#endif
}

void Plugin::setup(QImage &image, int &threadIndex,
				   QWaitCondition &bufferNotEmpty,
				   QWaitCondition &bufferNotFull, QMutex &pluginMutex,
				   QMutex &undoMutex, QVector<EventContainer *> &buffer,
				   int &numUsedBytes)
{
	m_stop = false;
	m_pImage = &image;
	m_indexThread = &threadIndex;
	m_pBufferNotEmpty = &bufferNotEmpty;
	m_pBufferNotFull = &bufferNotFull;
	m_pPluginMutex = &pluginMutex;
	m_pUndoMutex = &undoMutex;
	m_pBuffer = &buffer;
	m_pNumUsedBytes = &numUsedBytes;
}

QImage *Plugin::getImage() const{
	return m_pImage;
}
