#ifndef PLUGINS_H
#define PLUGINS_H

#include <QtGlobal>
#include <QList>
#include <QUuid>
#include <QThread>
#include <QRegion>
#include <QWaitCondition>
#include <QMutex>

#include "../definitions.h"
#include "../eventcontainer.h"

typedef struct PLUGININFO{
	QString shortName;
	quint32 version;
	QString versionString;
	QString description;
	QString author;
	QString authorEmail;
	QString copyright;
	QString homepage;
} PLUGININFO;

inline quint32 PLUGIN_VERSION(quint32 a, quint32 b, quint32 c, quint32 d){
	return (((a & 0xFF) << 24) | ((b & 0xFF) << 16) |
			((c & 0xFF) << 8) | (d & 0xFF) );
}

QString VERSION_TO_STRING(quint32 version);


class Plugin : public QThread {
	Q_PROPERTY(QString shortName WRITE setShortNameInfo
								 READ getShortNameInfo)
	Q_PROPERTY(quint32 version WRITE setVersionInfo
							   READ getVersionInfo)
	Q_PROPERTY(QString versionString READ getVersionStringInfo)
	Q_PROPERTY(QString description WRITE setDescriptionInfo
								   READ getDescriptionInfo)
	Q_PROPERTY(QString author WRITE setAuthorInfo
							  READ getAuthorInfo)
	Q_PROPERTY(QString authorEmail WRITE setAuthorEmailInfo
								   READ getAuthorEmailInfo)
	Q_PROPERTY(QString copyright WRITE setCopyrightInfo
								 READ getCopyrightInfo)
	Q_PROPERTY(QString homepage WRITE setHomepageInfo
								READ getHomepageInfo)
private:
	PLUGININFO m_pluginInfo;
protected:
	void setShortNameInfo(const QString &shortName);
	void setVersionInfo(const quint32 &version);
	void setDescriptionInfo(const QString &description);
	void setAuthorInfo(const QString &author);
	void setAuthorEmailInfo(const QString &authorEmail);
	void setCopyrightInfo(const QString &copyright);
	void setHomepageInfo(const QString &homepage);
	void setPluginInfo(const PLUGININFO &pluginInfo);
	void setPluginInfo(const QString &shortName,
					   const quint32 &version,
					   const QString &description,
					   const QString &author,
					   const QString &authorEmail,
					   const QString &copyright,
					   const QString &homepage);
public:
	QString getShortNameInfo() const;
	quint32 getVersionInfo() const;
	QString getVersionStringInfo() const;
	QString getDescriptionInfo() const;
	QString getAuthorInfo() const;
	QString getAuthorEmailInfo() const;
	QString getCopyrightInfo() const;
	QString getHomepageInfo() const;
	PLUGININFO getPluginInfo() const;
//-------------------------------------
//end of info part
private:
	bool awaitData();
	EventContainer *getData() const;
	void announceProcessingEnd();

	QWaitCondition *m_pBufferNotEmpty, *m_pBufferNotFull;
	QMutex *m_pPluginMutex, *m_pUndoMutex;
	QImage *m_pImage;
	QVector<EventContainer *> *m_pBuffer;
	int *m_pNumUsedBytes;
	int *m_indexThread;
	bool m_stop;

protected:
	void run();
	void waitForUndo();
	void releaseUndo();
	QImage *getImage() const;
	virtual QRegion onDataReceived(EventContainer *event) = 0;

public:
	Plugin();

	// NOTE: setup, stop and run could be made virtual to allow plugins overload
	// them, but we have to make sure it would actually make sense.
	void setup(QImage &image, int &threadIndex, QWaitCondition &bufferNotEmpty,
			  QWaitCondition &bufferNotFull, QMutex &pluginMutex,
			  QMutex &undoMutex, QVector<EventContainer *> &buffer,
			  int &numUsedBytes);
	void stop();

	virtual QList<QUuid> uuids() const = 0;

	virtual ~Plugin();
};

#endif // PLUGINS_H
