/*!
 * \file layer.h
 */

#ifndef LAYER_H
#define LAYER_H

#include <QWidget>
#include <QImage>
#include <QSize>
/*!
 * A layer class.
 * This class implements the layer. \n This represents a container for
 * the internal structure, which is not visible on the outside
 */
class Layer : public QWidget
{
    Q_OBJECT
	/*!
	 * @brief Holds the layer name
	 * @sa setName(), getName()
	 **/
	Q_PROPERTY(QString name READ getName WRITE setName)
private:
	QString m_name;
	QImage m_theImage;
public:
	/*!
	* Constructor of new empty layer.
	* Creates a new blank layer with dimensions given in the \a _width
	* and \a _height
	* @param _width width of the layer
	* @param _height height of the layer
	* @param parent pointer to the parent widget
	*/
	explicit Layer(uint _width, uint _height, QWidget *parent = 0);
	/*!
	* Constructor of new empty layer.
	* Creates a new blank layer with dimensions given in \a size
	* @param size size of the new layer
	* @param parent pointer to the parent widget
	*/
	explicit Layer(const QSize &size, QWidget *parent = 0);
	/*!
	* Constructor of new layer, whose content is determined.
	* Creates a layer from the given image, has the same dimension as
	* the given image.\n
	* Constructor used when opening the image from a disk, to create a layer 0
	* (background).
	* @param image image to be copied
	* @param parent pointer to the parent widget
	*/
	explicit Layer(const QImage &image, QWidget *parent = 0);


	/*!
	 * Returns the name of the layer
	 * @see setName()
	 * @return name of the layer.
	 */
	QString getName() const;

	/*!
	 * Set a new name of the layer
	 * @see getName()
	 * @param _name name to be set.
	 */
	void setName(QString _name);

	/*!
	 * Returns a pointer to the image after which you can paint
	 * @see getImage()
	 * @return pointer to the image
	 */
	QImage* getImage();

	/*!
	 * Creates a thumbnail of the designated size.
	 * The image is scaled to
	 * a rectangle as large as possible inside size, preserving
	 * the aspect ratio.
	 * @see createThumb()
	 * @param width width of generated thumbnail
	 * @param height height of generated thumbnail
	 * @return thumbnail image
	 */
	QImage createThumb(uint width, uint height) const;

};

#endif // LAYER_H
