/*!
 * \file layersdock.h
 */

#ifndef LAYERSDOCK_H
#define LAYERSDOCK_H

#include <QDockWidget>
#include <QListWidget>
#include <QListWidgetItem>
#include <QList>
#include <QToolButton>
#include <QAction>

#include "layer.h"
#include "layersmanager.h"

/*!
 * A layers dock widget class.
 * This class inherits from QDockWidget.\n The object of this class can be used
 * for visualization of the layers and to manage them, just pass a pointer to
 * an object of LayersManager class.
 */
class LayersDock : public QDockWidget {
    Q_OBJECT
private:
	QListWidget *m_pList;

	QToolButton *m_pAddButton;
	QToolButton *m_pRemoveButton;

	QAction *m_pAddAction;
	QAction *m_pRemoveAction;

	LayersManager *m_pLM;


	//methods
	/*!
	 * Populate the list with layers entries.
	 * @param layers list of pointers to layers
	 */
	void populateList(const QList<Layer*> &layers);
	/*!
	 * Calculate position of the items from the list to the right
	 * layer position, and vice versa
	 * @param pos position to convert, by default -1, if the parameter
	 *			  not given used is the current list row
	 * @return converted layer position
	 */
	inline int calculatePos( int pos = -1 );

public:
	/*!
	* Constructor of new dock.
	* Creates a new dock with given title and parent widget
	* @param title title of the dock (windowTitle)
	* @param parent pointer to the parent widget
	*/
	explicit LayersDock(const QString &title = QString(), QWidget *parent = 0);
	~LayersDock();

	/*!
	 * Set a pointer to layers manager which contents will be presented
	 * in the dock
	 * @see getLayersManager()
	 * @param lm pointer to layers manager object
	 */
	void setLayersManager(LayersManager *lm);
	/*!
	 * Returns pointer to layers manager which is currently set
	 * @see setLayersManager()
	 * @return pointer to layers manager object
	 */
	LayersManager* getLayersManager() const;

	/*!
	 * Set a new title of the dock
	 * @see getTitle()
	 * @param title title to be set
	 */
	void setTitle(const QString &title);
	/*!
	 * Returns the title of the dock
	 * @see setTitle()
	 * @return title of the dock
	 */
	QString getTitle() const;
private slots:
	/*!
	 * This slot is connected to Layers Manager object signal destroyed().
	 * When LM is destroing, this slot disconnect everything from object
	 * and set to safe state buttons and list in dock
	 * @param obj pointer to deleting object
	 */
	void childDestroyed( QObject *obj );
	/*!
	 * This slot is called when you click item in the list in dock
	 * @param item pointer to clicked item
	 */
	void itemActivated( QListWidgetItem *item );


	/*!
	 * This slot is called when you push add button in dock
	 */
	void addLayerPrivate();
	/*!
	 * This slot is called when you push remove button in dock
	 */
	void removeLayerPrivate();

public slots:
	/*!
	 * This slot is called when you want add something to the list of layers.
	 * In most cases is connected to LayersManager signal \a layerAdded
	 * @param pos position of new layer
	 * @param name name of the added layer
	 */
	void addLayer(uint pos, QString name);
	/*!
	 * This slot is called when you want remove something from the
	 * list of layers.
	 * In most cases is connected to LayersManager signal \a layerRemoved
	 * @param pos position of the layer to remove
	 */
	void removeLayer(uint pos);
	/*!
	 * This slot is called when you want rename layer in the list.
	 * In most cases is connected to LayersManager signal \a layerRenamed
	 * @param pos position of the layer to rename
	 * @param newName new name of the layer
	 */
	void renameLayer(uint pos, QString newName);

};

#endif // LAYERSDOCK_H
