#include "layersmanager.h"
#include <QFileInfo>
#include <QDebug>

LayersManager::LayersManager(const QString &fileName,
							 QWidget *parent, uint _width, uint _height,
							 QColor background) :
				QWidget(parent),
				m_layerIndex(0),
				m_activeLayer(0),
				m_width(_width),
				m_height(_height),
				m_transparent(true),
				m_background(background),
				//things to painting
				m_color(Qt::blue),
				m_thickness(3),
				m_antialiasing(true), m_alpha(255),
				m_numUsedBytes(0), m_index(0), m_indexThread(0),
				m_buffer(
					   QVector<EventContainer *>(INPUT_BUFFER_SIZE, NULL)),
				m_myTabletDevice(QTabletEvent::Stylus),
				m_lastPos(-1, -1),
				// initialize m_pLastEvent to NULL, so we can safely delete
				// it without any prior checks to see, whether it points to
				// a valid object or not.
				m_pLastEvent(NULL),
				m_undoTransaction(0),
				m_prevIdx(0),
				m_pUndoView(new QUndoView(&m_undoStack)),
				m_paintCnt(0)
{
	m_undoMutex.lock();
	//"binding" tools
	//if( (*(ToolsSettings::getInstance().bindInterface()))->type()
	//														== Plugin::Brush ){
	m_ppBrushInterface =
				reinterpret_cast<BrushInterface**>
				(ToolsSettings::getInstance().bindInterface());
	//}
	m_pBrush = ToolsSettings::getInstance().bindTool();

	connect(&ToolsSettings::getInstance(), SIGNAL(toolChanged(QUuid)), this,
			SLOT(toolChanged(QUuid)));


	//set background
	if( background.alpha() != 0){
		m_background = background;
		m_transparent = false;
		QPalette pal = palette();
		pal.setBrush(backgroundRole(), QBrush(Qt::white));
		setAutoFillBackground(true);
		setPalette(pal);
	}else{
		QPixmap pm(20, 20);
		QPainter pmp(&pm);
		pmp.fillRect(0, 0, 10, 10, Qt::white);
		pmp.fillRect(10, 10, 10, 10, Qt::white);
		pmp.fillRect(0, 10, 10, 10, Qt::lightGray);
		pmp.fillRect(10, 0, 10, 10, Qt::lightGray);
		pmp.end();
		QPalette pal = palette();
		pal.setBrush(backgroundRole(), QBrush(pm));
		setAutoFillBackground(true);
		setPalette(pal);
	}

	if( fileName.isEmpty() ){
		this->setMinimumSize(m_width, m_height);

		addLayer(0);
	}else{
		openImage( fileName );
	}

	setAttribute(Qt::WA_DeleteOnClose);

	connect(&m_undoStack, SIGNAL(indexChanged(int)), this,
			SLOT(undoIndexChanged(int)));

	m_pUndoView->setCleanIcon(
			QIcon(":/toolbar/res/toolbar/document-save.png") );

	if( (*m_ppBrushInterface) ){
		toolChanged( (*ToolsSettings::getInstance().bindTool()) );
	}
}

LayersManager::~LayersManager()
{
#ifdef THREAD_DEBUG_OUTPUT
	qDebug() << "awaiting for thread to finish.";
#endif
	m_undoMutex.unlock();
	stopPainting();
#ifdef THREAD_DEBUG_OUTPUT
	qDebug() << "thread has finished.";
#endif

	qDeleteAll(m_layers);
}

//-----------------------------------------------------------------------------
uint LayersManager::count() const{
	return m_layers.count();
}

uint LayersManager::size() const{
	return count();
}

QSize LayersManager::sizeHint() const{
	return QSize(m_width, m_height);
}

Layer* LayersManager::getLayer(uint pos) const{
	if( pos >= static_cast<uint>(m_layers.count()) )
		return NULL;
	return m_layers.at(pos);
}

QList<Layer*> LayersManager::getLayersList() const{
	return m_layers;
}

bool LayersManager::saveImage(const QString &fileName,
						 const char *fileFormat) const{
	QImage image( QSize(m_width, m_height),
				  QImage::Format_ARGB32_Premultiplied);

	if( !m_transparent )
		image.fill( m_background.rgba() );

	QPainter painter(&image);
	for( int i = 0; i < m_layers.count(); i++){
		if( m_layers.at(i)->isVisible() )
			painter.drawImage(0, 0, *(m_layers.at(i)->getImage()));
	}
	painter.end();

	QFileInfo fi(fileName);

	if( fi.suffix().isEmpty() && fileFormat == NULL )
		return image.save( fileName, "PNG" );

	return image.save( fileName, fileFormat );
}

bool LayersManager::openImage(const QString &fileName){
	QImage image;
	if( !image.load(fileName) )
		return false;

	addLayer(0, QString(), image);
	m_width = image.width();
	m_height = image.height();
	this->setMinimumSize(m_width, m_height);
	return true;
}

void LayersManager::renameLayer(uint pos, QString newName){
	m_layers.at( pos )->setName( newName );

	emit layerRenamed( pos, newName );
}

uint LayersManager::getActiveLayer() const{
	return m_activeLayer;
}

//-----------------------------------------------------------------------------
void LayersManager::addLayer(uint pos, QString _name, const QImage &image){
	Layer* layer;

	if( image.isNull() )
		layer = new Layer( m_width, m_height, this );
	else
		layer = new Layer(image, this);

	repaint();

	if( _name.isEmpty() ){
		layer->setName( QString(tr("Layer %1")).
						arg(QString::number(m_layerIndex++)) );
	}else{
		layer->setName( _name );
	}

	m_layers.insert( pos + 1, layer );

	emit layerAdded( pos, layer->getName() );
	setActiveLayer( pos );
}

void LayersManager::removeLayer(int pos){
	//i hope that pos is correct
	//if( pos >= m_layers.count() )
	//	return;
	m_layers.removeAt( pos );

	setActiveLayer( (pos==m_layers.count())?pos-1:pos );

	update();

	emit layerRemoved( pos );
}

void LayersManager::mergeLayers(const QList<uint> &/*_pos*/)
{

}

void LayersManager::changeOrder(uint start, uint destination){
	//start, destination are correct - i hope
	Layer* elem_start = m_layers.at( start );
	for( uint i = start; i < destination; i++ ){
		m_layers[i] = m_layers.at( i+1 );
	}

	m_layers.replace( destination, elem_start );
}

void LayersManager::changeVisible(uint pos){
	m_layers.at(pos)->setVisible( !m_layers.at(pos)->isVisible() );
}

void LayersManager::setActiveLayer(uint pos){
	m_activeLayer = pos;

	if( (*m_ppBrushInterface) != NULL ){
		toolChanged( (*ToolsSettings::getInstance().bindTool()) );
	}

	emit activeLayerChanged( m_layers.at(pos) );
}

void LayersManager::restart(){
	if ((*m_ppBrushInterface) != NULL) {
		stopPainting();
		toolChanged( (*ToolsSettings::getInstance().bindTool()) );
	}
}

//------------------------------------------------------------------------------
//things to painting

QUndoView* LayersManager::getUndoView() const{
	return m_pUndoView;
}

void LayersManager::stopPainting(bool wait, unsigned long time) {
	if ( (*m_ppBrushInterface) != NULL && (*m_ppBrushInterface)->isRunning() ){
		(*m_ppBrushInterface)->stop();
		if (wait) {
			(*m_ppBrushInterface)->wait(time);
		}
	}
}

void LayersManager::tabletEvent(QTabletEvent *event) {
	switch (event->type()) {
		case QEvent::TabletPress:
			if (*m_ppBrushInterface) {
				initiateValues();

				QTabletEvent *te = new QTabletEvent(event->type(), event->pos(),
									event->globalPos(), event->hiResGlobalPos(),
									event->device(), event->pointerType(),
									event->pressure(), event->xTilt(),
									event->yTilt(), event->tangentialPressure(),
									event->rotation(), event->z(),
									event->modifiers(), event->uniqueId());
				EventContainer *ec = new EventContainer(*m_pBrush, te);
				setupPainter(*ec);
				addEvent(ec);
			}
			break;
		case QEvent::TabletMove:
			if (*m_ppBrushInterface) {
				QTabletEvent *te = new QTabletEvent(event->type(), event->pos(),
									event->globalPos(), event->hiResGlobalPos(),
									event->device(), event->pointerType(),
									event->pressure(), event->xTilt(),
									event->yTilt(), event->tangentialPressure(),
									event->rotation(), event->z(),
									event->modifiers(), event->uniqueId());
				QList<QInputEvent *> list;
				list << te << m_pLastEvent;
				EventContainer *ec = new EventContainer(*m_pBrush, list);
				setupPainter(*ec);
				addEvent(ec);
			}
			break;
		case QEvent::TabletRelease:
			if (*m_ppBrushInterface) {
				QTabletEvent *te = new QTabletEvent(event->type(), event->pos(),
									event->globalPos(), event->hiResGlobalPos(),
									event->device(), event->pointerType(),
									event->pressure(), event->xTilt(),
									event->yTilt(), event->tangentialPressure(),
									event->rotation(), event->z(),
									event->modifiers(), event->uniqueId());
				EventContainer *ec = new EventContainer(*m_pBrush, te);
				setupPainter(*ec);
				addEvent(ec);
			}
			break;

		default:
			qWarning() <<
					"Sorry, your device and input type is not handled yet.";
	}
	m_pLastEvent = new QTabletEvent(event->type(), event->pos(),
									event->globalPos(), event->hiResGlobalPos(),
									event->device(), event->pointerType(),
									event->pressure(), event->xTilt(),
									event->yTilt(), event->tangentialPressure(),
									event->rotation(), event->z(),
									event->modifiers(), event->uniqueId());
}


void LayersManager::paintEvent(QPaintEvent * event) {
	QPainter painter(this);

	// NOTE: well, rectCount returns non-one only in very rare cases
	for( int i = 0; i < m_layers.count(); i++){
		QImage *img = m_layers.at(i)->getImage();
		for (int i = 0; i < event->region().rectCount(); i++) {
			painter.drawImage(QRect(event->region().rects().at(i).left(),
									event->region().rects().at(i).top(),
									event->region().rects().at(i).width(),
									event->region().rects().at(i).height()),
							  *img,
							  QRect(event->region().rects().at(i).left(),
									event->region().rects().at(i).top(),
									event->region().rects().at(i).width(),
									event->region().rects().at(i).height())
							  );
		}
	}
}

void LayersManager::mousePressEvent(QMouseEvent *event) {
	if (event->button() == Qt::LeftButton) {
		// NOTE: this should most probably substitute all m_theImage references,
		// once undo is restored.
		if (*m_ppBrushInterface) {
			initiateValues();

			//QPainter *painter = new QPainter(&m_theImage);
			QMouseEvent *me = new QMouseEvent(event->type(), event->pos(),
							event->globalPos(), event->button(),
							event->buttons(), event->modifiers());
			EventContainer *ec = new EventContainer(*m_pBrush, me);
			setupPainter(*ec);
			addEvent(ec);
		}

		m_lastPos = event->pos();
		m_pLastEvent = new QMouseEvent(event->type(), event->pos(),
									   event->globalPos(), event->button(),
									   event->buttons(), event->modifiers());
	}
}

void LayersManager::mouseMoveEvent(QMouseEvent *event) {
	if ((event->buttons() & Qt::LeftButton) && m_lastPos != QPoint(-1, -1)) {
		if (*m_ppBrushInterface) {
			//QPainter *painter = new QPainter(&m_theImage);
			QMouseEvent *me = new QMouseEvent(event->type(), event->pos(),
							event->globalPos(), event->button(),
							event->buttons(), event->modifiers());
			QList<QInputEvent *> list;
			list << me << m_pLastEvent;
			EventContainer *ec = new EventContainer(*m_pBrush, list);
			setupPainter(*ec);
			addEvent(ec);
		}
		m_lastPos = event->pos();
		m_pLastEvent = new QMouseEvent(event->type(), event->pos(),
									   event->globalPos(), event->button(),
									   event->buttons(), event->modifiers());
	}
}

void LayersManager::mouseReleaseEvent(QMouseEvent *event) {
	if (event->button() == Qt::LeftButton && m_lastPos != QPoint(-1, -1)) {
		if (*m_ppBrushInterface) {
			//QPainter *painter = new QPainter(&m_theImage);
			QMouseEvent *me = new QMouseEvent(event->type(), event->pos(),
							event->globalPos(), event->button(),
							event->buttons(), event->modifiers());
			EventContainer *ec = new EventContainer(*m_pBrush, me);
			setupPainter(*ec);
			addEvent(ec);
		}

		m_lastPos = QPoint(-1, -1);
		m_pLastEvent = new QMouseEvent(event->type(), event->pos(),
									   event->globalPos(), event->button(),
									   event->buttons(), event->modifiers());
	}
}


void LayersManager::addEvent(EventContainer *event) {
#ifdef THREAD_DEBUG_OUTPUT
	qDebug() << "addEvent start";
#endif
	m_pluginMutex.lock();
	if (m_numUsedBytes == INPUT_BUFFER_SIZE) {
		m_bufferNotFull.wait(&m_pluginMutex);
	}
	m_pluginMutex.unlock();

#ifdef THREAD_DEBUG_OUTPUT
	qDebug() << "addEvent start work";
#endif
	if (m_buffer.at(m_index) != NULL) {
		delete m_buffer.at(m_index);
	}
	m_buffer.replace(m_index, event);
	m_index = (m_index + 1) % INPUT_BUFFER_SIZE;
#ifdef THREAD_DEBUG_OUTPUT
	qDebug() << "addEvent end work";
#endif

	m_pluginMutex.lock();
	++m_numUsedBytes;
	m_bufferNotEmpty.wakeAll();
	m_pluginMutex.unlock();
#ifdef THREAD_DEBUG_OUTPUT
	qDebug() << "addEvent end";
#endif
}

// begins undo transaction which is cached until we commit it - we can save some
// memory by this approach
void LayersManager::beginUndoTransaction() {
#ifndef QT_NO_DEBUG
	//qDebug() << "begin:" << m_undoTransaction;
#endif
	// allow only one transaction at any given moment
	if (m_undoTransaction++ != 0) {
		return;
	}
	Q_ASSERT(m_undoTransaction > 0);
	//qDebug() << "begin ok";
	QImage *img = m_layers.at(m_activeLayer)->getImage();
	// create a deep copy of the image, as shallow-only copy will get its data
	// overwritten in another (plugin) thread.
	m_pTransactionImage = new QImage(img->copy());
	m_pUndoRegion = new QRegion();
	m_undoStack.beginMacro("Paint " + QString::number(m_paintCnt++));
	//qDebug() << "begin finished";
	m_undoMutex.unlock();
}

void LayersManager::appendUndoTransaction(const QRegion &region) {
#ifndef QT_NO_DEBUG
	//qDebug() << "appending to undo transaction:" << region;
#endif
	if (m_undoTransaction == 0) {
		return;
	}
	Q_ASSERT(m_undoTransaction > 0);
	// TODO: check, whether unite() just adds the region to list or rather
	// compresses them together, so that each pixel in region is included in
	// exactly one QRect.
	*m_pUndoRegion = m_pUndoRegion->united(
			region.intersected(m_pTransactionImage->rect()));
}

const UndoCommand *LayersManager::commitUndoTransaction() {
	m_undoMutex.lock();
#ifndef QT_NO_DEBUG
	//qDebug() << "commit:" << m_undoTransaction - 1;
#endif
	if (--m_undoTransaction > 0) {
		return NULL;
	}
	QImage *img = m_layers.at(m_activeLayer)->getImage();
	// blindly believe that there is no case where commit event could be fired
	// without triggering the begining of a transaction beforehand.
	Q_ASSERT(m_undoTransaction == 0);
	//qDebug() << "commit ok";
	// FIXME: check if QUndoStack handles deletion of this data in its dtor,
	// otherwise this here will leak.
	UndoCommand *uc = new UndoCommand(*m_pUndoRegion, *m_pTransactionImage,
									  *img, img);
	delete m_pTransactionImage;
	delete m_pUndoRegion;
	m_undoStack.push(uc);
	m_undoStack.endMacro();

	//report changes (for modified flag)
	emit imageChanged();

	return uc;
}

void LayersManager::initiateValues(){
	QVariant returnVal;
	returnVal =
			ToolsSettings::getInstance().getMemorySetting(Tool::AntiAliasing,
														  0);
	if( returnVal.isValid() )
		m_antialiasing = returnVal.toBool();

	returnVal =
			ToolsSettings::getInstance().getMemorySetting(Tool::SizeChange,
														  0);
	if( returnVal.isValid() )
		m_thickness = returnVal.toInt();

	returnVal =
			ToolsSettings::getInstance().getMemorySetting(Tool::ColorChange,
														  0);
	if( returnVal.isValid() )
		m_color = returnVal.value<QColor>();

	returnVal =
			ToolsSettings::getInstance().getMemorySetting(Tool::AlphaChange,
														  0);
	if( returnVal.isValid() )
		m_color.setAlpha( returnVal.toInt() );
}

void LayersManager::setupPainter(EventContainer &painter) {
	painter.setRenderHint(QPainter::Antialiasing, m_antialiasing);
	painter.setPen(QPen(m_color, m_thickness, Qt::SolidLine, Qt::RoundCap,
				   Qt::RoundJoin));

}

void LayersManager::undoIndexChanged(int idx) {
	//qDebug() << "uic: prev:" << m_prevIdx << ", idx:" << idx;
	int delta;
	if (m_prevIdx >= idx) {
		//qDebug() << "uic: undo";
		delta = -1;
	} else {
		//qDebug() << "uic: redo";
		idx--;
		delta = 1;
	}
	for (int i = m_prevIdx; i != idx + delta; i += delta) {
		// FIXME:
		// in current undo implementation we simply assume there is only and
		// exactly one child (we call QUndoStack::beginMacro(), QUS::push(),
		// QUS::endMacro(). this will *not* work for others
		//qDebug() << m_undoStack.command(i)->childCount();
		if (m_undoStack.count() && m_undoStack.command(i)->childCount()) {
			const QUndoCommand *quc = m_undoStack.command(i)->child(0);
			const UndoCommand *uc =	dynamic_cast<const UndoCommand *>(quc);
		// FIXME: should always be true for now (only one kind of UndoCommand)
			Q_ASSERT(uc);
			update(uc->region());
		}
	}
	m_prevIdx = idx;
}

void LayersManager::undo() {
	m_undoStack.undo();
}

void LayersManager::redo() {
	m_undoStack.redo();
}

void LayersManager::toolChanged(QUuid tool) {
	disconnect(this, SLOT(forceUpdate(QRegion)));
	disconnect(this, SLOT(appendUndoTransaction(QRegion)));
	disconnect(this, SLOT(beginUndoTransaction()));
	disconnect(this, SLOT(commitUndoTransaction()));
	stopPainting();
	qDebug() << "PaintArea::toolChanged(" << tool << ")";
	qDebug() << "*m_ppBrushInterface:" << (*m_ppBrushInterface);
	QImage *img = m_layers.at(m_activeLayer)->getImage();
	(*m_ppBrushInterface)->setup(*img, m_indexThread, m_bufferNotEmpty,
								 m_bufferNotFull, m_pluginMutex, m_undoMutex,
								 m_buffer, m_numUsedBytes);
	connect(*m_ppBrushInterface, SIGNAL(update(QRegion)), this,
			SLOT(forceUpdate(QRegion)));
	connect(*m_ppBrushInterface, SIGNAL(beginPainting()), this,
			SLOT(beginUndoTransaction()));
	connect(*m_ppBrushInterface, SIGNAL(endPainting()), this,
			SLOT(commitUndoTransaction()));
	connect(*m_ppBrushInterface, SIGNAL(update(QRegion)), this,
			SLOT(appendUndoTransaction(QRegion)));
	(*m_ppBrushInterface)->start();
}

void LayersManager::forceUpdate(QRegion region) {
	update(region);
}

