#ifndef LAYERSMANAGER_H
#define LAYERSMANAGER_H

#include <QWidget>
#include <QList>

#include "layer.h"

//things to painting
#include <QTabletEvent>
#include <QUndoStack>
#include <QUndoView>
#include <QDockWidget>
#include <QMainWindow>
#include <QWaitCondition>
#include <QMutex>

QT_BEGIN_NAMESPACE
class QColor;
class QImage;
class QPainterPath;
class QUndoView;
class QUuid;
QT_END_NAMESPACE

#include "undocommand.h"
#include "definitions.h"
#include "interfaces/brushinterface.h"
#include "toolssettings.h"
#include "eventcontainer.h"

class LayersManager : public QWidget
{
    Q_OBJECT
private:
	QList<Layer*> m_layers;
	uint m_layerIndex;

	uint m_activeLayer;

	uint m_width;
	uint m_height;

	bool m_transparent;

	QColor m_background;

	bool openImage(const QString &fileName);


public:
	explicit LayersManager(const QString &fileName = QString(),
						   QWidget *parent = 0, uint _width = 500,
						   uint _height = 300,
						   QColor background = Qt::transparent);
	~LayersManager();

	uint count() const;
	uint size() const;

	QSize sizeHint() const;

	Layer* getLayer(uint pos) const;
	QList<Layer*> getLayersList() const;

	bool saveImage(const QString &fileName,
				   const char *fileFormat = NULL) const;

	void renameLayer(uint pos, QString newName);

	uint getActiveLayer() const;
	void restart();

signals:
	void layerAdded(uint pos, QString name);
	void layerRemoved(uint pos);
	void layerRenamed(uint pos, QString newName);
	void activeLayerChanged(Layer* layer);

public slots:
	void addLayer(uint pos, QString _name = QString(),
				  const QImage &image = QImage());
	void removeLayer(int pos);
	void mergeLayers(const QList<uint> &_pos);
	void changeOrder(uint start, uint destination);
	void changeVisible(uint pos);
	void setActiveLayer(uint pos);

private slots:
	void beginUndoTransaction();
	void appendUndoTransaction(const QRegion &region);
	const UndoCommand *commitUndoTransaction();


//things to painting
public:
	QUndoView* getUndoView() const;
	void stopPainting(bool wait = true, unsigned long time = ULONG_MAX);
protected:
	void tabletEvent(QTabletEvent *event);
	void paintEvent(QPaintEvent *event);
	void mousePressEvent(QMouseEvent *event);
	void mouseMoveEvent(QMouseEvent *event);
	void mouseReleaseEvent(QMouseEvent *event);

private:
	void initiateValues();
	void setupPainter(EventContainer &painter);
	void addEvent(EventContainer *event);

	QColor m_color;
	int m_thickness;
	bool m_antialiasing;
	int m_alpha;

	QWaitCondition m_bufferNotEmpty;
	QWaitCondition m_bufferNotFull;
	QMutex m_pluginMutex, m_undoMutex;
	int m_numUsedBytes, m_index, m_indexThread;
	QVector<EventContainer *> m_buffer;

	QTabletEvent::TabletDevice m_myTabletDevice;

	// TODO: think how the versioned plugins can be implemented. One possible
	// way is to create a base interface e.g. "BrushInterface" which would be
	// inherited by all versions and {static, dynamic, qobject}_casted (I don't
	// know how to use them) to proper version. It might prove slow, though.

	//FIXME
	//need to be changed, think what happen if we use shapeInterface ;)
	BrushInterface **m_ppBrushInterface;

	const QUuid* m_pBrush;
	QPoint m_lastPos;
	QInputEvent *m_pLastEvent;
	int m_undoTransaction;
	QImage *m_pTransactionImage;
	QRegion *m_pUndoRegion;
	QUndoStack m_undoStack;
	int m_prevIdx;


	QUndoView *m_pUndoView;
	// NOTE: only for dev purposes
	int m_paintCnt;

public slots:
	void undo();
	void redo();
	void toolChanged(QUuid tool);
	void forceUpdate(QRegion region);

protected slots:
	void undoIndexChanged(int idx);
signals:
	void imageChanged();

};

#endif // LAYERSMANAGER_H
