#include "mdichild.h"

#include <QMessageBox>
#include <QApplication> //for qApp
#include <QFileInfo>
#include <QFileDialog>
#include <QRegExp>
#include <QStringList>

#include <QDebug>


MdiChild::MdiChild(QWidget *parent, QVariantMap *data) :
	QScrollArea(parent),
	m_wasModified(false),
	m_canClose(false),
	m_isUntitled(true),
	m_fileFormat("PNG"),
	m_pLM(NULL)
{
	setAttribute( Qt::WA_DeleteOnClose );
	setWindowFlags(this->windowFlags() | Qt::SubWindow );

	if( data != NULL )
		createNew( *data );
}

MdiChild::~MdiChild(){
	delete m_pLM;
}

/*---------------------------------------------------------------------------*/
void MdiChild::prepareToClose(bool _canClose){
	m_canClose = _canClose;
	this->close();
}

void MdiChild::closeEvent(QCloseEvent *event){
	if( m_canClose == false ){
		if( maybeSave() )
			emit closing(this);
		event->ignore();
	}
	if( m_canClose ){
		event->accept();
	}
}

/*---------------------------------------------------------------------------*/
LayersManager* MdiChild::getLayersManager() const{
	return m_pLM;
}

void MdiChild::createNew(const QVariantMap &data){
	this->setBackgroundRole( QPalette::Dark );
	this->setAttribute( Qt::WA_DeleteOnClose );

	if( data.value("open", QVariant(false)).toBool() )
		m_friendlyFileName = getFileName( data.value("fileName").toString() );
	else
		m_friendlyFileName = getFileName( data.value("windowName").toString() );

	setTitle( m_friendlyFileName );

	if( data.value("open", QVariant(false)).toBool() ){
		m_pLM = new LayersManager( data.value("fileName").toString(), this );


		setModified( false );
	}else{
		m_pLM = new LayersManager( QString(), this, data.value("width").toInt(),
								   data.value("height").toInt(),
								   data.value("background").value<QColor>());

		setModified( true );
	}

	this->setWidget( m_pLM );

	m_isUntitled = true;

	connect(m_pLM, SIGNAL(imageChanged()), this, SLOT(setModified()));
}

QString MdiChild::getTitle() const{
	if( m_wasModified )
		return this->windowTitle().replace("[*]","*");
	else
		return this->windowTitle().replace("[*]","");
}

/*---------------------------------------------------------------------------*/
bool MdiChild::maybeSave(){
	if ( m_wasModified ) {
		QMessageBox::StandardButton ret;
		ret = QMessageBox::warning(this, qApp->applicationName(),
					 tr("'%1' has been modified.\n"
						"Do you want to save your changes?")
					 .arg(m_friendlyFileName),
					 QMessageBox::Save | QMessageBox::Discard
					 | QMessageBox::Cancel);

		if ( ret == QMessageBox::Save )
			return save();
		else if ( ret == QMessageBox::Cancel )
			return false;
	}
	return true;
}

bool MdiChild::save(bool newFile){
	if( m_isUntitled || newFile){
		/*: default format and filter */
		QString filter = tr("PNG file (*.png);;");
		QString fileName =
				QFileDialog::getSaveFileName(this, tr("Save As"),
											 QString(),
				tr("PNG file (*.png);;JPG file (*.jpg);;"
				   "GIF file (*.gif);;Images (*.png *.gif *.jpg);;"), &filter);
		if( fileName.isEmpty() )
			return false;

		QRegExp rx("\\*\\.[\\w]{3}");
		rx.indexIn(filter);
		QStringList list = rx.capturedTexts();
		m_fileFormat = list.at(0).mid(2).toUpper();

		if( m_pLM->saveImage(fileName, m_fileFormat.toAscii().constData()) ){
			m_fileName = fileName;
			m_friendlyFileName = getFileName( fileName );
			setTitle( m_friendlyFileName );
			setModified( false );
			m_isUntitled = false;
			return true;
		}else{
			QMessageBox::warning(this, tr("Problem with saving"),
					tr("Could not write to a file, the image is not saved"));
			return false;
		}


	}

	if( m_wasModified ){
		if( m_pLM->saveImage(m_fileName, m_fileFormat.toAscii().constData()) ){
			setModified( false );
			return true;
		}else{
			QMessageBox::warning(this, tr("Problem with saving"),
					tr("Could not write to a file, the image is not saved"));
			return false;
		}

	}
	return false;
}

QString MdiChild::getFileName(const QString &file) const{
	return QFileInfo(file).fileName();
}

void MdiChild::setModified(bool _state){
	if( m_wasModified == _state )
		return;

	m_wasModified = _state;
	this->setWindowModified( _state );
	emit windowModifiedChanged( _state );
}

QString MdiChild::setTitle(QString _title){
	this->setWindowTitle( "[*]" + _title );
	emit windowTitleChanged( this->windowTitle() );
	return this->windowTitle();
}
