#include "toolssettings.h"

#include <QApplication>
#include <QMessageBox>
#include <QDebug>
#include <QVariant>
#include "definitions.h"


ToolsSettings::ToolsSettings(QObject *parent) :
	QObject(parent),
	m_pPlugin(NULL),
	m_memoryDBconnection("memoryDB")
{
	createConnection();

}

ToolsSettings::~ToolsSettings(){
}

bool ToolsSettings::initiateDatabase(){

	QSqlQuery query;

	query.exec("PRAGMA foreign_keys = ON;");

	if ( query.lastError().isValid() ){
		 qWarning() << "0" << query.lastError();
		 return false;
	}

	query.exec("CREATE TABLE tools "
			   "(id INTEGER CONSTRAINT tools_key PRIMARY KEY, "
			   "tool VARCHAR(36) NOT NULL CONSTRAINT tools_uniq UNIQUE);");

	if ( query.lastError().isValid() ){
		 qWarning() << "1" << query.lastError();
		 return false;
	}

	query.exec("CREATE TABLE settings "
			   "(id INTEGER CONSTRAINT settings_key PRIMARY KEY, "
			   "setting INTEGER NOT NULL CONSTRAINT settings_uniq UNIQUE);");

	if ( query.lastError().isValid() ){
		 qWarning() << "2" << query.lastError();
		 return false;
	}


	query.exec("CREATE TABLE settingsvalues "
			   "(id INTEGER CONSTRAINT settingsval_key PRIMARY KEY, "
			   "tool INTEGER NOT NULL REFERENCES tools(id), "
			   "setting INTEGER NOT NULL REFERENCES settings(id), "
			   "device INTEGER NOT NULL, "
			   "value BLOB NOT NULL);");

	if ( query.lastError().isValid() ){
		 qWarning() << "3" << query.lastError();
		 return false;
	}


	return true;
}

bool ToolsSettings::createConnection(){
	QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE");
	db.setDatabaseName( "toolssettings.sqlite" );
	if (!db.open()) {
		QMessageBox::critical(0, tr("Cannot open database"),
			tr("Unable to establish a database connection.\n"
			   "Click Cancel to exit."), QMessageBox::Ok );
		qWarning("file database");
		qApp->closeAllWindows();
		qApp->exit( ERROR_SQL_CONNECTION );
		return false;
	}

	if( db.tables().count() == 0 && !initiateDatabase() ){
			qCritical()<<"Cannot create tables";
	}



	QSqlDatabase memoryDB =
			QSqlDatabase::addDatabase("QSQLITE", m_memoryDBconnection);

	memoryDB.setDatabaseName(":memory:");
	if (!memoryDB.open()) {
		QMessageBox::critical(0, tr("Cannot open database"),
			tr("Unable to establish a database connection.\n"
			   "Click Cancel to exit."), QMessageBox::Ok );
		qWarning("memory database");
		qApp->closeAllWindows();
		qApp->exit( ERROR_SQL_CONNECTION );
		return false;
	}

	QSqlQuery query(memoryDB);

	query.exec("CREATE TABLE settings "
			   "(id INTEGER CONSTRAINT settingsval_key PRIMARY KEY, "
			   "setting INTEGER NOT NULL,  "
			   "device INTEGER NOT NULL, "
			   "value BLOB NOT NULL);");
	if ( query.lastError().isValid() ){
		 qWarning() << "memory query" << query.lastError();
		 return false;
	}

	return true;
}


QVariant ToolsSettings::getSetting(const QString &tool, int device,
						uint feature) const{
qDebug()<<"getFile"<<tool<<device<<feature;
	QSqlQuery query;
	query.prepare("SELECT value FROM settingsvalues WHERE device=? "
				  "AND tool=(SELECT id FROM tools WHERE tool=? ) "
				  "AND setting=(SELECT id FROM settings WHERE setting=? );");
	query.addBindValue( device );
	query.addBindValue( tool );
	query.addBindValue( feature );
	if( query.exec() && query.isSelect() ){
		if( query.first() && query.isValid() )
			return query.value( 0 );
	}

	//if not exist return empty
	return QVariant();
}


bool ToolsSettings::setSetting(const QString &_tool, int device,
							   uint feature, QVariant value){
	QString tool;
	if( _tool.isEmpty() )
		tool = m_tool.toString();
	else
		tool = _tool;
qDebug()<<"setFile"<<tool<<device<<feature<<value;
	QSqlQuery query;
	query.prepare("UPDATE settingsvalues "
				  "SET value=? "
				  "WHERE device=? "
				  "AND tool=(SELECT id FROM tools WHERE tool=? ) "
				  "AND setting=(SELECT id FROM settings WHERE setting=? );");
	query.addBindValue( value );
	query.addBindValue( device );
	query.addBindValue( tool );
	query.addBindValue( feature );
	if( query.exec() && query.numRowsAffected() > 0 ){
		return true;
	}

	//something is missing
	query.prepare("INSERT INTO tools VALUES(null, ?);");
	query.addBindValue( tool );
	if( query.exec() && query.numRowsAffected() ){
		this->setSetting(tool, device, feature, value);
		return true;
	}

	query.prepare("INSERT INTO settings VALUES(null, ?);");
	query.addBindValue( feature );
	if( query.exec() && query.numRowsAffected() ){
		this->setSetting(tool, device, feature, value);
		return true;
	}

	query.prepare("INSERT INTO settingsvalues "
				  "VALUES(null, (SELECT id FROM tools WHERE tool=? ), "
				  "(SELECT id FROM settings WHERE setting=? ), "
				  "?, ?);");
	query.addBindValue( tool );
	query.addBindValue( feature );
	query.addBindValue( device );
	query.addBindValue( value );
	if( query.exec() && query.numRowsAffected() == 0 ){
		qWarning() << "Something goes wrong, data not added to database";
		qWarning() << "tool:"<< tool << "feature:" << feature
				<< "device:" << device << "value:" << value;
		qWarning() << query.lastQuery();
		qWarning() << query.lastError();
		return false;
	}
	return true;
}


QVariant ToolsSettings::getMemorySetting(uint feature, int /*device*/) const{
qDebug()<<"getMemory"<<0<<feature;
	QSqlQuery query(QSqlDatabase::database(m_memoryDBconnection));
	query.prepare("SELECT value FROM settings WHERE device=? "
				  "AND setting=?;");
	query.addBindValue( 0 /*device*/ );
	query.addBindValue( feature );
	if( query.exec() && query.isSelect() ){
		if( query.first() && query.isValid() )
			return query.value( 0 );
	}

	return QVariant();
}

bool ToolsSettings::setMemorySetting(uint feature,
									 int /*device*/, QVariant value){
qDebug()<<"setMemory"<<feature<<value;

	QSqlQuery query(QSqlDatabase::database(m_memoryDBconnection));
	query.prepare("UPDATE settings "
				  "SET value=? "
				  "WHERE device=? "
				  "AND setting=?;");
	query.addBindValue( value );
	query.addBindValue( 0 /*device*/ );
	query.addBindValue( feature );
	if( query.exec() && query.numRowsAffected() > 0 ){
		return true;
	}


	//if row not exist
	query.prepare("INSERT INTO settings "
				  "VALUES(null, ?, ?, ?);");
	query.addBindValue( feature );
	query.addBindValue( 0 /*device*/ );
	query.addBindValue( value );
	if( query.exec() && query.numRowsAffected() == 0 ){
		qWarning() << "Something goes wrong, data not added to memory database";
		qWarning() << "feature:" << feature
				<< "device:" << 0 /*device*/ << "value:" << value;
		qWarning() << query.lastQuery();
		qWarning() << query.lastError();
		return false;
	}
	return true;
}

bool ToolsSettings::wipeSettings(bool safe){
qDebug()<<"wipeSettings";
	if( safe ){
		QSqlQuery query(QSqlDatabase::database(m_memoryDBconnection));
		return query.exec("DELETE FROM settings");
	}
	return false;
}

void ToolsSettings::setTool(Plugin *_plugin, QUuid _tool){
	qDebug()<<"setTool"<<_plugin<<_tool;
	m_pPlugin = _plugin;
	m_tool = _tool;
	emit toolChanged(m_tool);
}

const QUuid* ToolsSettings::bindTool() const{
	return &m_tool;
}

Plugin** ToolsSettings::bindInterface(){
	return &m_pPlugin;
}
