#include "undocommand.h"

#include <QImage>

#ifndef QT_NO_DEBUG
	#include <QDebug>
#endif

#define USE_ZLIB

UndoCommand::UndoCommand(const QRegion region, const QImage before,
						 const QImage after, QImage *img) :
		m_region(region), m_targetImage(img), m_depth(img->depth()),
		m_first(true) {
	QByteArray bytesBeforeBuffer, bytesAfterBuffer;
	for (int i = 0; i < m_region.rects().size(); i++) {
		const QImage tempImgB = before.copy(m_region.rects().at(i));
		const QImage tempImgA = after.copy(m_region.rects().at(i));
		bytesBeforeBuffer.append((const char*)tempImgB.bits(),
								 tempImgB.byteCount());
		bytesAfterBuffer.append((const char*)tempImgA.bits(),
								tempImgA.byteCount());
	}
#ifdef USE_ZLIB
	//qDebug() << "using zlib";
	//long sizeBefore, sizeAfter;
	//sizeBefore = bytesBeforeBuffer.size() + bytesAfterBuffer.size();
	m_bytesBefore = qCompress(bytesBeforeBuffer); // use zlib's default level
	m_bytesAfter = qCompress(bytesAfterBuffer);
	//sizeAfter = m_bytesBefore.size() + m_bytesAfter.size();
	//qDebug() << "size before:" << sizeBefore << ", sizeAfter:" << sizeAfter <<
	//		", ratio:" << (double)sizeAfter / sizeBefore;
#else
	//qDebug() << "NOT using zlib";
	m_bytesBefore = bytesBeforeBuffer;
	m_bytesAfter = bytesAfterBuffer;
#endif
}

void UndoCommand::undo() {
#ifndef QT_NO_DEBUG
	qDebug() << "undo";
#endif
	uchar *imgBits = m_targetImage->bits();
	// amount of bytes blited
	int b = 0;
#ifdef USE_ZLIB
	QByteArray undoData = qUncompress(m_bytesBefore);
#endif
	for (int i = 0; i < m_region.numRects(); i++) {
		int startX = m_region.rects().at(i).topLeft().x();
		int startY = m_region.rects().at(i).topLeft().y();
		int width = m_region.rects().at(i).width();
		int height = m_region.rects().at(i).height();
		Q_ASSERT(m_targetImage);
		// FIXME: won't work if the image has changed its depth or size
		for (int y = startY; y < startY + height; y++) {
			for (int x = startX; x < startX + width; x++) {
				for (int d = 0; d < m_depth / 8; d++) {
					int idx = (y * m_targetImage->width() + x) *
							  (m_depth / 8) + d;
#ifdef USE_ZLIB
					imgBits[idx] = undoData.at(b);
#else
					imgBits[idx] = m_bytesBefore.at(b);
#endif
					b++;
				}
			}
		}
	}
}

// FIXME: Provide custom UndoStack.
// redo is called unnecessarily by QUndoStack - we paint "directly" to screen
// and cache the UndoCommand, which when commited, has executed redo() on it
void UndoCommand::redo() {
#ifndef QT_NO_DEBUG
	qDebug() << "redo";
#endif
	// a workaround for QUndoStack's need to execute redo() on commands being
	// push onto a stack. Follow KISS principle.
	if (m_first) {
		m_first = false;
		return;
	}
	uchar *imgBits = m_targetImage->bits();
	// amount of bytes blited
	int b = 0;
#ifdef USE_ZLIB
	QByteArray redoData = qUncompress(m_bytesAfter);
#endif
	for (int i = 0; i < m_region.numRects(); i++) {
		int startX = m_region.rects().at(i).topLeft().x();
		int startY = m_region.rects().at(i).topLeft().y();
		int width = m_region.rects().at(i).width();
		int height = m_region.rects().at(i).height();
		Q_ASSERT(m_targetImage);
		// FIXME: won't work if the image has changed its depth or size
		for (int y = startY; y < startY + height; y++) {
			for (int x = startX; x < startX + width; x++) {
				for (int d = 0; d < m_depth / 8; d++) {
					int idx = (y * m_targetImage->width() + x) *
							  (m_depth / 8) + d;
#ifdef USE_ZLIB
					imgBits[idx] = redoData.at(b);
#else
					imgBits[idx] = m_bytesAfter.at(b);
#endif
					b++;
				}
			}
		}
	}
}

const QRegion UndoCommand::region() const {
	return m_region;
}
