#include "xmlparser.h"

#include "definitions.h"
#include <QDir>
#include <QStringList>
#include <QFile>
#include <QXmlStreamAttributes>
#include <QXmlStreamAttribute>
#include <QRegExp>
#include <QDateTime>

#ifndef QT_NO_DEBUG
	#include <QDebug>
#endif

XMLParser::XMLParser(QObject *parent) :
    QObject(parent)
{
}

bool XMLParser::readPresets(VariantMapList &data){
	QDir dir(QDir::current().path() + "/presets");

	QStringList files = dir.entryList( QDir::Files );

	foreach(QString fileName,files){
		QFile file( dir.absoluteFilePath(fileName) );
#ifndef QT_NO_DEBUG
		qDebug()<<file.fileName();
#endif
		if( file.open(QIODevice::ReadOnly | QIODevice::Text) ){
			parsePreset( file, data );
		}
	}
	if( data.isEmpty() )
		return false;
	return true;

}

bool XMLParser::parsePreset(QFile &file, VariantMapList &data){
	QXmlStreamReader reader(&file);
	QVariantMap content;
	while(!reader.atEnd() && !reader.hasError()){
		reader.readNext();

		if(reader.isStartDocument()) {
			continue;
		}

		if( reader.isStartElement() &&
			reader.name() == QString::fromUtf8("preset") )
			continue;

		if( reader.isStartElement() ){
			QXmlStreamAttributes attrList = reader.attributes();
			foreach(QXmlStreamAttribute attr, attrList){
				content.insert(
						reader.name().toString() +
						"_" + attr.qualifiedName().toString(),
						attr.value().toString() );
			}
			content.insert(
					reader.name().toString(), reader.readElementText() );
		}

	}

	if( reader.hasError() ){
#ifndef QT_NO_DEBUG
		qDebug()<<file.fileName()<<":error:"<<reader.error()<<
				reader.errorString();
#endif

		reader.clear();
		return false;
	}
	//if no error, add read data to list
	data.append(content);


	reader.clear();
	return true;
}

bool XMLParser::savePreset(const QVariantMap &data){
	QString name = data.value("name","defaultName").toString().toLower();
	name = name.replace(QRegExp("[\\W]+"),"_");

	QString fileName = "preset-" + name + ".xml";

	QFile file(QDir::current().path() + "/presets/" + fileName);

	if( file.exists() ){
		uint time = QDateTime::currentDateTime().toTime_t();
		fileName = "preset-" + name + QString::number(time) + ".xml";
		file.setFileName(QDir::current().path() + "/presets/" + fileName);
	}

	if( file.open(QIODevice::WriteOnly | QIODevice::Text)){
		QXmlStreamWriter writer(&file);
		writer.setAutoFormatting(true);//enable indent
		writer.setAutoFormattingIndent(-1);//indent 1 tab
		writer.setCodec("UTF-8");
		writer.writeStartDocument("1.0"); //xml version

		writer.writeStartElement(QString::fromUtf8("preset"));

		QStringList keys = data.keys();
		QStringList keyAttr = keys.filter("_");
		cleanKeys(keys, keyAttr);
		StringMap attrs = mapAttributes(keyAttr);

		foreach(QString key, keys){
			writer.writeStartElement( key );
			QStringList attrList = attrs.values( key );
			foreach(QString attr, attrList){
				writer.writeAttribute(attr,
									  data.value(key + "_" + attr).toString());
			}
			writer.writeCharacters(data.value( key ).toString());
			writer.writeEndElement();
		}

		writer.writeEndElement();

		writer.writeEndDocument();
		return true;
	}
	return false;
}


void XMLParser::cleanKeys(QStringList &source, const QStringList &toDelete){
	foreach(QString elem, toDelete){
		int pos = elem.indexOf("_");
		if( !source.contains( elem.left(pos) ) )
			source.append( elem.left(pos) );

		source.removeAll( elem );
	}
	source.sort();
}

StringMap XMLParser::mapAttributes(const QStringList &keyAttr){
	StringMap attributes;
	foreach(QString elem, keyAttr){
		int pos = elem.indexOf("_");
		attributes.insertMulti(elem.left(pos),elem.right(elem.size()-pos-1));
	}
	return attributes;
}
